// AP(r) Computer Science Marine Biology Simulation:
// The MBSDataFileHandler class is copyright(c) 2002 College Entrance
// Examination Board (www.collegeboard.com).
//
// This class is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation.
//
// This class is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

import java.io.File;
import java.io.FileReader;
import java.io.LineNumberReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.StringTokenizer;

/**
 *  AP&reg; Computer Science Marine Biology Simulation:<br>
 *  An <code>MBSDataFileHandler</code> object reads and writes information
 *  about an environment and the objects in it to and from a data file.
 *  An environment is a conceptual grid (whether it is actually represented
 *  that way or not), and the objects in the environment are
 *  <code>Locatable</code> objects that occupy (row, column) positions
 *  in the grid.
 *
 *  <p>
 *  The <code>MBSDataFileHandler</code> object assumes that the environment
 *  objects are ones that the Marine Biology Simulation program can act on,
 *  and that they have constructors that expect three parameters: the
 *  environment in which the object will act, its initial location, and its
 *  initial direction.
 *
 *  <p>
 *  The first line of the input file should be either
 *  <pre>
 *     bounded <i>rows columns</i>
 *  </pre>
 *  or
 *  <pre>
 *     unbounded
 *  </pre>
 *  where <i>rows</i> and <i>columns</i> are integers indicating the
 *  number of rows and columns in a bounded environment. 
 *
 *  Each line after the first represents an object in the environment
 *  and should be in the following format:
 *  <pre>
 *     class row-pos col-pos direction
 *  </pre>
 *  where <code>class</code> is a string indicating the class name of
 *  the object ("Fish" for example), <code>row-pos</code> and
 *  <code>col-pos</code> are integers indicating its row and column
 *  position, and <code>direction</code> is a compass direction (either
 *  a word like North or Northeast or a number of degrees).
 *
 *  <p>
 *  The <code>MBSDataFileHandler</code> class is
 *  copyright&copy; 2002 College Entrance Examination Board
 *  (www.collegeboard.com).
 *
 *  @author Alyce Brady
 *  @author Scot Drysdale
 *  @author Julie Zelenski
 *  @version 1 August 2002
 *  @see Environment
 *  @see BoundedEnv
 *  @see UnboundedEnv
 *  @see Locatable
 *  @see MBSException
 *  @see MBSFactory
 **/
public class MBSDataFileHandler
{
    // Encapsulated data used to read/write environment info from a file
    private File myFile;   		  	        // configuration data file
    private LineNumberReader inputReader;  	// buffered input w/ line number
    private StringTokenizer tokenizer;   	// parses tokens from a line


    /** Constructs an MBSDataFileHandler object to read and/or write
     *  environment information to and from a file.
     *  @param  file       java.io.File object from which to read or write
     **/
    public MBSDataFileHandler(File file)
    {
        myFile = file;
    }

    /** Reads information about an environment from an initial configuration
     *  data file.
     *  @param  boundedEnvClass    class to use when constructing a bounded
     *                             environment
     *  @param  unboundedEnvClass  class to use when constructing an unbounded
     *                             environment
     *  @throws java.io.FileNotFoundException if file cannot be opened
     *  @throws MBSException   if invalid information is read from file
     **/
    public Environment readMBSEnv(Class boundedEnvClass,
                                  Class unboundedEnvClass)
        throws java.io.FileNotFoundException
    {
        // Open the file for reading.
        inputReader = new LineNumberReader(new FileReader(myFile));

        // Read environment information and construct environment.
        Environment env = buildEnv(boundedEnvClass, unboundedEnvClass);

        // Read MBS environment objects until there are no more.
        while ( readEnvObject(env) )
            ;

        return env;
    }

    /** Reads environment specifications and constructs environment.
     *  @param  boundedEnvClass    class to use when constructing a bounded
     *                             environment
     *  @param  unboundedEnvClass  class to use when constructing an unbounded
     *                             environment
     *  @throws MBSException  if invalid environment information is read
     **/
    private Environment buildEnv(Class boundedEnvClass, Class unboundedEnvClass)
    {
        // Read in environment type.
        String envType;
        try
        {
            envType = readString();
        }
        catch (Exception e)
        { 
            throw new MBSException("Error reading environment type (line " 
                                   + inputReader.getLineNumber() + ")"); 
        }

        if ( envType.equals("bounded") )
        {
            int envRows, envCols;

            // Read (and validate) environment dimensions.
            try
            {
                envRows = readInt();
                envCols = readInt();
            }
            catch (Exception e)
            { 
                throw new MBSException("Error reading environment dimensions (line "
                                       + inputReader.getLineNumber() + ")"); 
            }

            if ( envRows <= 0 || envCols <= 0 )
                throw new MBSException("Environment dimensions must be positive (line "
                                       + inputReader.getLineNumber() + ")");

            // Construct the appropriate bounded environment.
            return MBSFactory.constructEnv(boundedEnvClass, envRows, envCols);
        }
        else if ( envType.equals("unbounded") )
        {
            // Construct the appropriate unbounded environment.
            return MBSFactory.constructEnv(unboundedEnvClass);
        }
        else
            throw new MBSException("File must begin with bounded or unbounded (line "
                                   + inputReader.getLineNumber() + ")"); 
    }

    /** Reads the next environment object and constructs it, using
     *  the <code>MBSFactory</code> class. (The factory must know how to
     *  construct environment objects with the class name read in from
     *  the file.)  The environment object adds itself to the environment.
     *  @param env   the environment in which object should be created
     *  @return <code>true</code> if object was successfully read,
     *          <code>false</code> at EOF
     *  @throws MBSException   if invalid object information is read
     **/
    private boolean readEnvObject(Environment env)
    {
        String className = null;    // class name of env object read
        Class cls;                  // class for env object read
        Location loc;         	    // location of env object read
        Direction dir;       	    // direction env object is facing

        try
        {
            // Read the class name first (a null string means end of file).
            className = readString();
            if ( className == null )
            {
                // If end-of-file encountered, close file.
                inputReader.close();
                return false;
            }
            cls = Class.forName(className);

            // Now read the object's location and direction.
            loc = readLocation(env);
            dir = readDirection();
            
            // Construct the object in the environment at the specified location
            // and direction.
            MBSFactory.constructEnvObject(cls, env, loc, dir);
        }
        catch (ClassNotFoundException e)
        {
              throw new MBSException("Cannot find class named "
                                     + className + " (line " 
                                     + inputReader.getLineNumber() + ")");
        }
        catch (java.io.IOException e)
        {
            // Convert i/o exceptions to MBSException.
            throw new MBSException("Error reading environment object (line " 
                                   + inputReader.getLineNumber() + ")"); 
        }
        catch (MBSException e)
        {
            throw new MBSException("Error constructing object (line " 
                                   + inputReader.getLineNumber() + ")"); 
        }

        return true;
    }

    /** Reads in a Location object (must be a valid location in
     *  <code>env</code>).
     *  @param env      environment in which location must be valid
     *  @return  the newly created Location object
     *  @throws MBSException   if invalid location information is read
     **/
    private Location readLocation(Environment env)
    {
        int row, col;
        try
        {
            // Read the location.
            row = readInt();
            col = readInt();
        }
        catch (Exception e)
        {
            // Convert reading exceptions to MBSException.
            throw new MBSException("Error reading location (line " 
                                   + inputReader.getLineNumber() + ")"); 
        }
        Location loc = new Location(row, col);
        
        // Verify that location is valid in this environment.
        if ( env.isValid(loc) )
            return loc;
        else
            throw new MBSException("Location " +  loc +
                                   " is not valid in this environment (line " 
                                   + inputReader.getLineNumber() + ")"); 
    }

    /** Reads in a new Direction object.  The direction could be either an
     *  integer or a named compass direction recognized by the Direction
     *  constructor.
     *  @return the newly created direction object
     *  @throws MBSException   if invalid direction information read
     **/
    private Direction readDirection()
        throws java.io.IOException
    {
        // Direction could be either an integer or a named compass
        // direction.
        String dirName = null;
        Direction dir = null;
        
        try
        {	// Attempt to interpret token as direction in degrees.
            dirName = readString();
            int degrees = Integer.parseInt(dirName);
            dir = new Direction(degrees);
        } 
        catch (Exception e)
        {
            try 
            {
            	// If that didn't work, try to interpret as a named direction.
                dir = new Direction(dirName);
            }
            catch (Exception e2)
            {
                throw new MBSException("Error reading direction (line " 
                                       + inputReader.getLineNumber() + ")");
            }
        }
        
        return dir;
    }

    /** Returns the next token in the file as a string.
     *  @return     a String containing the next token in the file; or null
     *              if end of file is encountered
     *  @throws     java.io.IOException if an input error occurs
     **/
    private String readString()
        throws java.io.IOException
    {
        // Read in a new line if there are no more tokens in current line.
        while ( tokenizer == null || ! tokenizer.hasMoreTokens() )
        {
            String line = inputReader.readLine();

            // Did we encounter end of file?
            if ( line == null )
                return null;
            tokenizer = new StringTokenizer(line);
        }

        // Return next token.
        return tokenizer.nextToken();
    }

    /** Returns the next token in the file as an integer.
     *  @return    an int containing the next number in the file
     *  @throws    java.io.EOFException if EOF
     *  @throws    java.lang.NumberFormatException if next token is not an int
     *  @throws    java.io.IOException if another type of input error occurs
     **/
    private int readInt()
        throws java.io.IOException
    {
        // Read in number as string, then convert to integer.
        String token = readString();
        if ( token == null )
            throw new java.io.EOFException();
        return Integer.parseInt(token);
    }

    /** Writes information about an environment into a data file.
     *  As currently implemented, it assumes that all objects in an
     *  environment are instances of the Fish class or a Fish subclass.
     *  If the Marine Biology Simulation is extended to introduce a
     *  new superclass that represents objects in the environment (such
     *  as MarineAnimal or EnvironmentObject), this method should be
     *  modified to cast the objects to that class rather than to Fish.
     *  @param  env    environment to write to file
     *  @throws java.io.FileNotFoundException if file cannot be opened
     *  @throws java.io.IOException if error writing to file
     **/
    public void writeMBSEnv(Environment env)
        throws java.io.IOException
    {
        PrintWriter out = new PrintWriter(new FileWriter(myFile));

        // Save environment type (and dimensions if necessary).
        if (env.numRows() == -1 || env.numCols() == -1)
            out.println("unbounded");
        else
            out.println("bounded " + env.numRows() + " " + env.numCols());
            
        // Save all MBS environment objects (class, location, and direction).
        Locatable[] objList = env.allObjects();
        for ( int index = 0; index < objList.length; index++ )
        {
            Fish fish = (Fish)objList[index];
            String className = fish.getClass().getName();
            int row = fish.location().row();
            int col = fish.location().col();
            String dir = fish.direction().toString();
            out.println(className + " " + row + " " + col + " " + dir);
        }
        out.close();
    }

}
