// AP(r) Computer Science Marine Biology Simulation:
// The MBSFactory class is copyright(c) 2002 College Entrance
// Examination Board (www.collegeboard.com).
//
// This class is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation.
//
// This class is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

import java.lang.reflect.Constructor;
import java.awt.Color;
import java.util.Set;
import java.util.TreeSet;
import java.util.Comparator;

/**
 *  AP&reg; Computer Science Marine Biology Simulation:<br>
 *  The <code>MBSFactory</code> class provides a set of static methods 
 *  for constructing objects in the Marine Biology Simulation program,
 *  such as fish and environments, as well as tracking the available
 *  classes to use in constructing those objects.
 *
 *  <p>
 *  The <code>MBSFactory</code> class is
 *  copyright&copy; 2002 College Entrance Examination Board
 *  (www.collegeboard.com).
 *
 *  @author Julie Zelenski
 *  @version 1 August 2002
 **/
public class MBSFactory
{
    // Static (class) variables maintained by the factory hold the sets
    // of known environment object, BoundedEnv, and UnboundedEnv classes.
    // The comparator used for the TreeSets will keep the sets sorted by
    // class name.
    private static Comparator classCmp = new Comparator() {
        public int compare(Object o1, Object o2) { 
            return o1.toString().compareTo(o2.toString());
        }};
    private static Set envObjectClasses = new TreeSet(classCmp);
    private static Set boundedEnvClasses = new TreeSet(classCmp);
    private static Set unboundedEnvClasses = new TreeSet(classCmp);
    
    // Class constants that define the parameter types for common MBS constructors
    private static final Class[] TWO_ARG_TYPES = {Environment.class, Location.class};
    private static final Class[] THREE_ARG_TYPES = {Environment.class, Location.class, Direction.class};
    private static final Class[] FOUR_ARG_TYPES = {Environment.class, Location.class, Direction.class, Color.class};
    private static final Class[] BOUNDED_ARGS = {int.class, int.class};
    private static final Class[] UNBOUNDED_ARGS = null;
    

    /** Creates an object of the given class.
     *  @param   cls        	 class of new object
     *  @param   parameterTypes  parameter types expected by constructor
     *                           for class
     *  @param   parameters      actual parameters to pass to constructor
     *  @return  the newly created object
     *  @throws MBSException   if an object of the specified class cannot be
     *                         constructed with the specified parameters
     **/
    public static Object constructObject(Class cls, Class[] parameterTypes,
                                         Object[] parameters)                                        
    {
       Object newObject = null;        // the new object
       
        try 
        {
            // Construct an instance via class constructor.
            if ( parameterTypes == null || parameters == null )
            {
                // if no parameters; use default constructor
                newObject = cls.newInstance();
            }
            else
            {
               	// use the constructor that matches the parameterTypes
                Constructor objCons = cls.getConstructor(parameterTypes);
                newObject = objCons.newInstance(parameters);
            }
            return newObject;
        }
        catch (Exception e)
        {
            throw new MBSException("Cannot construct " + cls.getName() + 
                                   " object due to " + e);
        }
     }
     

    /** Creates an instance of the given environment object class using a
     *  two-argument constructor that takes an environment and a location.
     *  @param   cls  	class of new environment object
     *  @param   env 	environment in which this object will act
     *  @param   loc  	location the object will occupy
     *  @return  the newly created environment object
     *  @throws MBSException   if an object of the specified class cannot be
     *                         constructed with the specified parameters
     **/
    public static Object constructEnvObject(Class cls, Environment env,
                                            Location loc)
    {
        Object[] parameters = {env, loc};
        return constructObject(cls, TWO_ARG_TYPES, parameters);
    }

    /** Creates an instance of the given environment object class using a
     *  three-argument constructor that takes an environment, a location,
     *  and a direction.
     *  @param   cls  	class of new environment object
     *  @param   env 	environment in which this object will act
     *  @param   loc  	location the object will occupy
     *  @param   dir   	direction the object will face
     *  @return  the newly created environment object
     *  @throws MBSException   if an object of the specified class cannot be
     *                         constructed with the specified parameters
     **/
    public static Object constructEnvObject(Class cls, Environment env,
                                            Location loc, Direction dir)
    {
        Object[] parameters = {env, loc, dir};
        return constructObject(cls, THREE_ARG_TYPES, parameters);
    }
    
    /** Creates an instance of the given environment object class using a
     *  four-argument constructor that takes an environment, a location,
     *  a diredction, and a color.
     *  @param   cls  	class of new environment object
     *  @param   env 	environment in which this object will act
     *  @param   loc  	location the object will occupy
     *  @param   dir   	direction the object will face
     *  @param   color 	color of the object
     *  @return  the newly created environment object
     *  @throws MBSException   if an object of the specified class cannot be
     *                         constructed with the specified parameters
     **/
    public static Object constructEnvObject(Class cls, Environment env, 
                                            Location loc, Direction dir,
                                            Color color)
    {
        Object[] parameters = {env, loc, dir, color};
        return constructObject(cls, FOUR_ARG_TYPES, parameters);
    }
    
    /** Creates an instance of an Environment using the default constructor
     *  of the given class. Used to create unbounded environments; no 
     *  dimensions are specified.
     *  @param   cls  	class of new environment
     *  @return  the newly created environment
     *  @throws MBSException   if an environment of the specified class cannot be
     *                         constructed
     **/
    public static Environment constructEnv(Class cls)
    {
        return (Environment)constructObject(cls, UNBOUNDED_ARGS, null);
    }

    /** Creates an instance of an Environment using the 2-argument constructor
     *  of the given class. Used to create bounded environments;  
     *  dimensions must be specified.
     *  @param   cls  	class of new environment
     *  @return  the newly created environment
     *  @throws MBSException   if an environment of the specified class cannot be
     *                         constructed with the given number of rows and columns
     **/
    public static Environment constructEnv(Class cls, int numRows, int numCols)
    {
        Object[] parameters = {new Integer(numRows), new Integer(numCols)};
        return (Environment)constructObject(cls, BOUNDED_ARGS, parameters);
    }


    /** Returns the set of environment object classes known to the factory.  
     *  Classes are added to the factory via the
     *  <code>addEnvObjClassNames</code> method.
     *  @return  the set of environment object classes
     **/
    public static Set envObjectClasses() 
    { 
        return envObjectClasses; 
    }
    
    /** Returns the set of bounded environment classes known to the factory.  
     *  Classes are added to the factory via the <code>addBoundedClassNames</code> method.
     *  @return  the set of bounded environment classes
     **/
    public static Set boundedEnvClasses() 
    {
        return boundedEnvClasses; 
    }
    
    /** Returns the set of unbounded environment classes known to the factory.  
     *  Classes are added to the factory via the <code>addUnboundedClassNames</code> method.
     *  @return  the set of unbounded environment classes
     **/
    public static Set unboundedEnvClasses() 
    {
        return unboundedEnvClasses; 
    }
    
    /** Helper to add new classes to the factory. Classes are specified by
     *  string name. For each name, error-checking is done to ensure the name
     *  class exists, is accessible, and has the proper constructor. If all checks
     *  out, the class is added to the appropriate set maintained by the factory.
     *  On errors, a message is printed and the class is skipped.
     *  @param  classNames an array of class names
     *  @param  whichCategory  descriptive name for category 
     *                         (e.g., "bounded environment")
     **/
    protected static void addClassesToFactory(String[] classNames, 
                                              String whichCategory)
    {
        for (int i = 0; i < classNames.length; i++) 
        {
            String errStart = "Discarding " + whichCategory + " choice \"" + 
                              classNames[i] + "\" because ";
            try 
            {
                Class cls = Class.forName(classNames[i]);
                if (whichCategory.equals("bounded environment"))
                {
                    if (isValidEnvClass(cls, BOUNDED_ARGS))
                        boundedEnvClasses.add(cls);
                }
                else if (whichCategory.equals("unbounded environment"))
                {
                    if (isValidEnvClass(cls, UNBOUNDED_ARGS))
                        unboundedEnvClasses.add(cls);
                }
                else 
                {
                    envObjectClasses.add(cls);
                }
            }
            catch (ClassNotFoundException e)
            {
                System.err.println(errStart + "no class found with that name.");
            } 
            catch (ClassCastException e)
            {
                System.err.println(errStart + e.getMessage());
            }
            catch (NoSuchMethodException e) 
            {
                System.err.println(errStart + "it doesn't have the proper constructor.");
            } 
        }
    }

   
    /** Adds environment object classes to the set maintained by the factory.
     *  If the named class cannot be found or the class is not a proper
     *  Locatable object, an error message is printed and the class is not
     *  added to the set.
     *  @param  classNames an array of environment object class names
     **/
    public static void addEnvObjClassNames(String[] classNames)
    {
        addClassesToFactory(classNames, "environment object");
    }
        
    /** Adds bounded environment classes to the set maintained by the factory. 
     *  If the named class cannot be found or the class is not a proper bounded
     *  environment, an error message is printed, and that class is not added to the set.
     *  @param  classNames an array of bounded environment class names
     **/
    public static void addBoundedEnvClassNames(String[] classNames)
    {
        addClassesToFactory(classNames, "bounded environment");
    }


    /** Adds unbounded environment classes to the set maintained by the factory. 
     *  If the named class cannot be found or the class is not a proper unbounded
     *  environment, an error message is printed, and that class is not added to the set.
     *  @param  classNames an array of unbounded environment class names
     **/
    public static void addUnboundedEnvClassNames(String[] classNames)
    {
        addClassesToFactory(classNames, "unbounded environment");
    }
    
    /** Verifies that a class has the required constructor and
      *  is properly assignable to Environment.  This is used to vet
      *  environment classes given to the factory before
      *  adding them to the list of known classes.
     **/
    public static boolean isValidEnvClass(Class cls,
                                          Class[] ctorParameters)
    throws NoSuchMethodException
    {	
        return hasCorrectCtor(cls, Environment.class, ctorParameters);
    }
    
    /** Verifies that a class has the required constructor and
      *  is properly assignable to Locatable.  This is used to vet
      *  environment object classes given to the factory before
      *  adding them to the list of known classes.
     **/
    public static boolean isValidEnvObjectClass(Class cls,
                                                Class[] ctorParameters)
    throws NoSuchMethodException
    {	
        return hasCorrectCtor(cls, Locatable.class, ctorParameters);
    }
    
    /** Verifies that a class has the required constructor and
      *  is properly assignable. This is used to vet fish/environment classes
      *  given to the factory before adding them to the list of known classes.
     **/
    protected static boolean hasCorrectCtor(Class cls, Class requiredCls,
                                            Class[] ctorParameters)
    throws NoSuchMethodException
    {	
            if (!requiredCls.isAssignableFrom(cls)) 
                throw new ClassCastException("not compatible with " + requiredCls + ".");
            Constructor ctor = cls.getConstructor(ctorParameters);
            return true;
    }
    
    
    /** Reports wehther a given environment object class has an accessible
     *  four-arg constructor that takes Environment, Location, Direction,
     *  and Color.
     **/
    public static boolean hasFourArgCtor(Class cls)
    {
        try 
        {
            return isValidEnvObjectClass(cls, FOUR_ARG_TYPES);
        }
        catch (NoSuchMethodException e) {}
        return false;
    }

}
